/*! ********************************************
 * MultiTimberSynthBase.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth;

import sora.graybox.GrayBox;
import sora.graybox.IGrayBoxModule;
import sora.graybox.MidiEvent;
import sora.graybox.ProcessData;
//import java.util.function.Consumer;
public abstract class MultiTimberSynthBase implements IGrayBoxModule
{
// FunctionalInterface ---
	//@FunctionalInterface
	//protected interface Function_RPN<T,U,V,W> { void func(T t, U u, V v, W w); }
	//@FunctionalInterface
	//protected interface Function_CC<T> { void func(T t); }
	@FunctionalInterface
	protected interface Function_RPN{ void func(int ch, int param, int mask, int inc); }
	@FunctionalInterface
	protected interface Function_CC{ void func(MidiEvent ev); }
	@FunctionalInterface
	protected interface Function_PRC{ void func(float[] outL, float[] outR, int end); }
// -----
protected abstract class VoiceUnitBase
{
// Tables
	// Volume, Pan
	protected float[] _mainVolTable = null;
	protected float[] _exprVolTable = null;
	protected float[] _veloVolTable = null;
	protected float[] _panTable     = null;
// ptr
	//final Function_PRC[] _PROC_PTRS = new Function_PRC[16];
	Function_PRC process = null;
// Synthesizer Data
	protected float _masterVolume = 0.0f;
// VoiceUnit Status
	/*
	- VoiceUnit State -
				[ FEDC BA98 7654 3210 ] 16Bits Flag.
		value -> [ ---- ---- -KKK dcba ] "-" = Reserved Bit.
		(All Bits "0" = VoiceUnit is DeActive.)
		
		- Operator State -
					[ 3210 ]
			value -> [ dcba ]
			a = Operator 0
			b = Operator 1
			c = Operator 2
			d = Operator 3
			
		- VoiceUnit Mode - (Exclusive BitFlag)
					[ FEDC BA98 7654 ]
			value -> [ ---- ---- -KKK ]
				-> [ ---- ---- -cba ]
			
			a = Key On
			b = Key Off
			c = Sonud Off
	*/
	int state      = 0;
	int partNumber = -1;
	int noteNumber = 0;
	int startFrame = 0;  // Offset Start Sample Frame
	// Volume
	//protected float[] _volumeTable = null;
	protected int _prPan = 64;     // preset Pan   (0-127)
	protected float _vPanL = 0.0f; // total pan L      (val = _panTable[      _prPan + Part.ccPan])
	protected float _vPanR = 0.0f; // total pan R      (val = _panTable[256 - _prPan - Part.ccPan])
	protected float _vPtVol= 0.0f; // CC Volume        (val = _mainVolTable[Part.volume])
	protected float _vPtExp= 0.0f; // CC Expression    (val = _exprVolTable[Part.expression])
	protected float _vVel  = 0.0f; // velocity         (val = _veloVolTable[velocity])
	protected float _vVol  = 0.0f; // etc              (val = _vVel * masterVolume)
	protected float _vOut  = 0.0f; // output volume    (val = _vPtVol * _vPtExp * _vVol)
	protected float _vOutL = 0.0f; // output volume L  (val = _vOut * _vPanL)
	protected float _vOutR = 0.0f; // output volume R  (val = _vOut * _vPanR)
	
	void killProcess()
	{
		state      = 0;
		partNumber = -1;
	}
	abstract void keyOff();
	abstract void soundOff();
	void keyOn(int ch, int note, int velo) {}
	void setMasterVolume(float val)
	{
		_masterVolume = val;
		if (state == 0) return;
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	}
	void updPartBendSens(int val){}
	void updPartPitchBend(int val){}
	void updPartModulation(int val){}
	void updPartVolume(int val)
	{
		_vPtVol = _mainVolTable[val];
		_vOut = _vPtVol * _vPtExp * _vVol;
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	}
	void updPartExpression(int val)
	{
		_vPtExp = _exprVolTable[val];
		_vOut = _vPtVol * _vPtExp * _vVol;
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	}
	void updPartPan(int val)
	{
		_vPanL = _panTable[      _prPan + val];
		_vPanR = _panTable[256 - _prPan - val];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	}
// Constructor
	VoiceUnitBase()
	{
		//_PROC_PTRS[0] = this::_proc0;
		//_PROC_PTRS[1] = this::_proc1;
		//_PROC_PTRS[2] = this::_proc2;
		//_PROC_PTRS[3] = this::_proc3;
		//_PROC_PTRS[4] = this::_proc4;
		//_PROC_PTRS[5] = this::_proc5;
		//_PROC_PTRS[6] = this::_proc6;
		//_PROC_PTRS[7] = this::_proc7;
		//_PROC_PTRS[8] = this::_proc8;
		//_PROC_PTRS[9] = this::_proc9;
		//_PROC_PTRS[10] = this::_procA;
		//_PROC_PTRS[11] = this::_procB;
		//_PROC_PTRS[12] = this::_procC;
		//_PROC_PTRS[13] = this::_procD;
		//_PROC_PTRS[14] = this::_procE;
		//_PROC_PTRS[15] = this::_procF;

	}
// porc
	//protected void _proc0(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _proc1(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _proc2(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _proc3(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _proc4(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _proc5(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _proc6(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _proc7(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _proc8(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _proc9(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _procA(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _procB(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _procC(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _procD(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _procE(float[] outL, float[] outR, int end){ killProcess(); }
	//protected void _procF(float[] outL, float[] outR, int end){ killProcess(); }
}
// -----
// Factory -----
	protected abstract void _recreateVU(int maxPoly);
	protected void _resetKeyParams(){}
// Tables -----
	protected float[] _masterVolumeTable = null;  // for sysEx, NRPN
	protected void _resetVolumeTables()
	{
		if (_masterVolumeTable == null)_masterVolumeTable = new float[128];
		for (int i = 0; i < 128; i++)
		{
			// 0 to -96dB
			_masterVolumeTable[i] = (float)(Math.pow(2.0, (i + 1) / 8.0) / 65536.0);
		}
	}
// Synthesizer - Data -----
	public void factoryDataReset(){}
// Synthesizer - Main -----
	// Synth Status
	public int midiInPorts = 0xFFFF;
	public int midiOutPorts = 0xFFFF;
	public boolean receiveSysEx = false;
	protected byte _manufacID = 0x7D;
	protected byte _deiceID = 0x10;
	protected byte _modelID = 0;
	protected float _sampleRate = 44100.0f;
	protected float _referencePitch = 440.0f;
	protected float _masterVolume = 1.0f;
	// Channel Paramters
	public VoicePart[] voiceParts = null;
	protected int _maxPMN = 0;
	// VoiceUnits
	protected int _maxPolyphony = 0;
	protected VoiceUnitBase[] _voiceUnits = null;
	protected int _nton = 0;
	protected int _ntonM = 0;
	// getter, setter ---
	public float getMasterVolume() { return _masterVolume; }
	public void setMasterVolume(float val)
	{ 
		_masterVolume = val;
		for (int i = 0; i < _maxPolyphony; i++)_voiceUnits[i].setMasterVolume(_masterVolume);
	}
	public float getSampleRate(){ return _sampleRate; }
	public void setSampleRate(float val){ _sampleRate = val; _resetKeyParams(); }
	public float getReferencePitch(){ return _referencePitch; }
	public void setReferencePitch(float val){ _referencePitch = val; _resetKeyParams(); }
	public int getPolyphony()
	{
		int _cnt = 0;
		for (int i = 0; i < _maxPolyphony; i++) if (_voiceUnits[i].state != 0)_cnt++;
		return _cnt;
	}
	public int getMaxPolyphony(){ return _maxPolyphony; }
	public int setMaxPolyphony(int num)
	{
		if (num < 1 || num == _maxPolyphony) return _maxPolyphony;
		if (num >= 1024) num = 1024;
		else for (int v = 1; v < 1024; v <<= 1) if (v >= num){ num = v; break; } 
		_nton = 0;
		_ntonM = num - 1;
		_recreateVU(num);
		_resetKeyParams();
		return _maxPolyphony;
	}
	public byte getManufacID(){ return _manufacID; }
	public void setManufacID(byte id){ _manufacID = (byte)(id & 127); }
	public byte getDeviceID(){ return _deiceID; }
	public void setDeviceID(byte id){ _deiceID = (byte)(id & 127); }
	public byte getModelID(){ return _modelID; }
	public void setModelID(byte id){ _modelID = (byte)(id & 127); }
// Constructor
	protected MultiTimberSynthBase(){ this(32, 440.0f, 1.0f); }
	protected MultiTimberSynthBase(int maxPoly, float refPitch, float masVol)
	{
		_referencePitch = refPitch;
		_masterVolume = masVol;
	// Table ---
		_resetVolumeTables();
	// RPN ---
		for (int i = 0; i < 256; i++)_RPN_FUNC[i] = this::_rpnFunc_UB;
		_RPN_FUNC[0x100] = this::_rpnFunc_Null;
		// rpn
		_RPN_FUNC[0] = this::_rpnFunc_00;
		_RPN_FUNC[1] = this::_rpnFunc_01;
		_RPN_FUNC[2] = this::_rpnFunc_02;
		// nrpn
		_RPN_FUNC[0x80] = this::_nrpnFunc_00;
		_RPN_FUNC[0x80 | 0x20] = this::_nrpnFunc_20;
		_RPN_FUNC[0x80 | 0x21] = this::_nrpnFunc_21;
	// CC ---
		for (int i =   0; i < 120; i++)_CC_FUNC[i] = this::_ccFunc_channelVoiceUB;
		for (int i = 120; i < 128; i++)_CC_FUNC[i] = this::_ccFunc_channelModeUB;
		_CC_FUNC[0x01] = this::_ccFunc_01;
		_CC_FUNC[0x06] = this::_ccFunc_06;
		_CC_FUNC[0x07] = this::_ccFunc_07;
		_CC_FUNC[0x0A] = this::_ccFunc_0A;
		_CC_FUNC[0x0B] = this::_ccFunc_0B;
		_CC_FUNC[0x26] = this::_ccFunc_26;
		_CC_FUNC[0x60] = this::_ccFunc_60;
		_CC_FUNC[0x61] = this::_ccFunc_61;
		_CC_FUNC[0x62] = this::_ccFunc_62;
		_CC_FUNC[0x63] = this::_ccFunc_63;
		_CC_FUNC[0x64] = this::_ccFunc_64;
		_CC_FUNC[0x65] = this::_ccFunc_65;
		_CC_FUNC[0x78] = this::_ccFunc_78;
		_CC_FUNC[0x79] = this::_ccFunc_79;
		_CC_FUNC[0x7B] = this::_ccFunc_7B;
	// Part ---
		voiceParts = new VoicePart[16];
		for (int i = 0; i < 16; i++) voiceParts[i] = new VoicePart(1);
	// VoiceUnit ---
		if (maxPoly >= 1024) maxPoly = 1024;
		else for (int v = 1; v < 1024; v <<= 1) if (v >= maxPoly){ maxPoly = v; break; } 
		_nton = 0;
		_ntonM = maxPoly - 1;
		_maxPolyphony = maxPoly;
	}
// Controls ---
	public void initialize()
	{
		_resetVolumeTables();
		allSoundStop();
		allPartReset(true);
		factoryDataReset();
		_rpnCtrlFlag = true;
		_nrpnCtrlFlag = false;
		receiveSysEx = false;
	}
	public int getDefaultPartMode(int part) { return voiceParts[part].defaultMode; }
	public void setDefaultPartMode(int part, int mode) { if (mode > _maxPMN)mode = _maxPMN; voiceParts[part].defaultMode = mode; }
	public int getPartMode(int part) { return voiceParts[part].mode; }
	public void setPartMode(int part, int mode)
	{
		if (mode > _maxPMN) mode = _maxPMN;
		if (mode == voiceParts[part].mode) return;
		voiceParts[part].mode = mode;
		allSoundStop(part);
	}
	public void allPartReset(){ allPartReset(false); }
	public void allPartReset(boolean modeReset)
	{
		for (int i = 0; i < 16; i++) voiceParts[i].reset();
		if (modeReset)for (int i = 0; i < 16; i++)voiceParts[i].mode = voiceParts[i].defaultMode;
	}
	public void allSoundStop(){ allSoundStop(-1); }
	public void allSoundStop(int part)
	{
		if (part < 0) for (int i = 0; i < _maxPolyphony; i++)_voiceUnits[i].killProcess();
		else for (int i = 0; i < _maxPolyphony; i++)if (_voiceUnits[i].partNumber == part)_voiceUnits[i].killProcess();
	}
	public void allSoundOff(){ allSoundOff(-1); }
	public void allSoundOff(int part)
	{
		if (part < 0) for (int i = 0; i < _maxPolyphony; i++)_voiceUnits[i].soundOff();
		else for (int i = 0; i < _maxPolyphony; i++)if (_voiceUnits[i].partNumber == part)_voiceUnits[i].soundOff();
	}
// MidiCtrl ---
	protected MidiEvent _evBuff = new MidiEvent();
	public void noteOff(int ch, int key)
	{
		//_evBuff.type = 0x80;
		_evBuff.channel = ch;
		_evBuff.value0 = key;
		_noteOff(_evBuff);
	}
	public void noteOn(int ch, int key, int velo)
	{
		if (voiceParts[ch].mode == 0) return;
		//_evBuff.type = 0x90;
		_evBuff.channel = ch;
		_evBuff.value0 = key;
		_evBuff.value1 = velo;
		if (velo > 0)_noteOn(_evBuff);
		else _noteOff(_evBuff);
	}
	public void keyAfterTouch(int ch, int key, int val)
	{
		//_evBuff.type = 0xA0;
		_evBuff.channel = ch;
		_evBuff.value0 = key;
		_evBuff.value1 = val;
		_keyAfterTouch(_evBuff);
	}
	public void controleChange(int ch, int cc, int val)
	{
		//_evBuff.type = 0xB0;
		_evBuff.channel = ch;
		//_evBuff.value0 = cc;
		_evBuff.value1 = val;
		_CC_FUNC[cc].func(_evBuff);
	}
	public void programChange(int ch, int num)
	{
		//_evBuff.type = 0xC0;
		_evBuff.channel = ch;
		_evBuff.value0 = num;
		_programChange(_evBuff);
	}
	public void channelAfterTouch(int ch, int val)
	{
		//_evBuff.type = 0xD0;
		_evBuff.channel = ch;
		_evBuff.value0 = val;
		_channelAfterTouch(_evBuff);
	}
	public void pitchBendChange(int ch, int lsb, int msb)
	{
		//_evBuff.type = 0xE0;
		_evBuff.channel = ch;
		_evBuff.value0 = lsb;
		_evBuff.value1 = msb;
		_pitchBendChange(_evBuff);
	}
	// val[0 - 16383]
	public void pitchBendChangeW(int ch, int val)
	{
		//_evBuff.type = 0xE0;
		_evBuff.channel = ch;
		_evBuff.value0 = val & 0x7f;
		_evBuff.value1 = val >> 7 & 0x7f;
		_pitchBendChange(_evBuff);
	}
	public void sysExMessage(byte[] msg)
	{
		if (!receiveSysEx) return;
		//_evBuff.type = 0xF0;
		_evBuff.channel = 0;
		_evBuff.data = msg;
		_readSysExData(_evBuff);
		_evBuff.data = null;
	}
// MidiEvent ---
	protected void _noteOff(MidiEvent ev)
	{
		int _key = ev.value0, _ch = ev.channel;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			// KeyOn -->
			if ((_vu.state & 0x010) != 0) if (_vu.partNumber == _ch && _vu.noteNumber == _key)_vu.keyOff();
		}
	}
	protected void _noteOn(MidiEvent ev)
	{
		final int _start = _nton;
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitBase _vu = _voiceUnits[_nton];
			if (_vu.state == 0)
			{
				_vu.startFrame = ev.offsetFrames;
				_vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
			
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitBase _vu = _voiceUnits[_nton];
			// (KeyOff || SoundOff) -->
			if ((_vu.state & 0x60) != 0)
			{
				_vu.startFrame = ev.offsetFrames;
				_vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}

		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitBase _vu = _voiceUnits[_nton];
			// KeyOn -->
			if ((_vu.state & 0x10) != 0)
			{
				_vu.startFrame = ev.offsetFrames;
				_vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
	}
	protected void _keyAfterTouch(MidiEvent ev){}
	protected void _programChange(MidiEvent ev)
	{
		int _ch = ev.channel;
		voiceParts[_ch].programChange = ev.value0;
		if (voiceParts[_ch].mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++) if (_voiceUnits[i].partNumber == _ch)_voiceUnits[i].soundOff();
	}
	protected void _channelAfterTouch(MidiEvent ev)
	{
		voiceParts[ev.channel].chPressure = ev.value0;
	}
	protected void _pitchBendChange(MidiEvent ev)
	{
		int _ch = ev.channel;
		VoicePart _pt = voiceParts[_ch];
		_pt.pitchBend = ev.value1 << 7 | ev.value0;
		if (_pt.mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartPitchBend(_pt.pitchBend);
		}
	}
	protected void _readSysExData(MidiEvent ev) {}
// MidiEvent - CC -----
	// Function Detail : function(MidiEvent ev)
	//protected final Function_CC<MidiEvent>[] _CC_FUNC = new Function_CC[128];
	protected final Function_CC[] _CC_FUNC = new Function_CC[128];
	private void _ccFunc_channelVoiceUB(MidiEvent ev)
	{
		voiceParts[ev.channel].ccValues[ev.value0] = ev.value1;
	}
	private void _ccFunc_channelModeUB(MidiEvent ev){}
	private void _ccFunc_01(MidiEvent ev)
	{
		// Modulation --->
		int _ch = ev.channel;
		VoicePart _pt = voiceParts[_ch];
		_pt.ccValues[1] = ev.value1;
		if (_pt.mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartModulation(ev.value1);
		}
	}
	private void _ccFunc_06(MidiEvent ev)
	{
		// Data Entry MSB --->
		voiceParts[ev.channel].ccValues[6] = ev.value1;
		_RPN_FUNC[voiceParts[ev.channel].paramNumber].func(ev.channel, ev.value1 << 7, 0x7f, 0);
	}
	private void _ccFunc_07(MidiEvent ev)
	{
		// Volume --->
		int _ch = ev.channel;
		VoicePart _pt = voiceParts[_ch];
		_pt.ccValues[7] = ev.value1;
		if (_pt.mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartVolume(ev.value1);
		}
	}
	private void _ccFunc_0A(MidiEvent ev)
	{
		// Pan --->
		int _ch = ev.channel;
		VoicePart _pt = voiceParts[_ch];
		_pt.ccValues[10] = ev.value1;
		if (_pt.mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartPan(ev.value1);
		}
	}
	private void _ccFunc_0B(MidiEvent ev)
	{
		// Expression --->
		int _ch = ev.channel;
		VoicePart _pt = voiceParts[_ch];
		_pt.ccValues[11] = ev.value1;
		if (_pt.mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartExpression(ev.value1);
		}
	}
	private void _ccFunc_26(MidiEvent ev)
	{
		// Data Entry LSB --->
		voiceParts[ev.channel].ccValues[38] = ev.value1;
		_RPN_FUNC[voiceParts[ev.channel].paramNumber].func(ev.channel, ev.value1, 0x3f80, 0);
	}
	private void _ccFunc_60(MidiEvent ev)
	{
		// Data Inc --->
		_RPN_FUNC[voiceParts[ev.channel].paramNumber].func(ev.channel, 0, 0x3fff, 1);
	}
	private void _ccFunc_61(MidiEvent ev)
	{
		// Data Dec --->
		_RPN_FUNC[voiceParts[ev.channel].paramNumber].func(ev.channel, 0, 0x3fff, -1);
	}
	private void _ccFunc_62(MidiEvent ev)
	{
		// NRPN LSB --->
		if (!_nrpnCtrlFlag) return;
		VoicePart _pt = voiceParts[ev.channel];
		_pt.ccValues[98] = ev.value1;
		_pt.nrpn = 0x4000 | _pt.nrpn & 0xff80 | ev.value1;
		if (_pt.nrpn >= 0xc000)  // Received MSB,LSB
		{
			_pt.paramNumber = _pt.nrpn & 0x3fff;
			if (_pt.paramNumber <= 0x7f)_pt.paramNumber |= 0x80;
			else _pt.paramNumber = 0x100;
		}
	}
	private void _ccFunc_63(MidiEvent ev)
	{
		// NRPN MSB --->
		if (!_nrpnCtrlFlag) return;
		VoicePart _pt = voiceParts[ev.channel];
		_pt.ccValues[99] = ev.value1;
		_pt.nrpn = 0x8000 | _pt.nrpn & 0xc07f | ev.value1 << 7;
		if (_pt.nrpn >= 0xc000)  // Received MSB,LSB
		{
			_pt.paramNumber = _pt.nrpn & 0x3fff;
			if (_pt.paramNumber <= 0x7f)_pt.paramNumber |= 0x80;
			else _pt.paramNumber = 0x100;
		}
	}
	private void _ccFunc_64(MidiEvent ev)
	{
		// RPN LSB --->
		if (!_rpnCtrlFlag) return;
		VoicePart _pt = voiceParts[ev.channel];
		_pt.ccValues[100] = ev.value1;
		_pt.rpn = 0x4000 | _pt.rpn & 0xff80 | ev.value1;
		if (_pt.rpn >= 0xc000)  // Received MSB,LSB
		{
			_pt.paramNumber = _pt.rpn & 0x3fff;
			if (_pt.paramNumber > 0x7f)_pt.paramNumber = 0x100;
		}
	}
	private void _ccFunc_65(MidiEvent ev)
	{
		// RPN MSB --->
		if (!_rpnCtrlFlag) return;
		VoicePart _pt = voiceParts[ev.channel];
		_pt.ccValues[101] = ev.value1;
		_pt.rpn = 0x8000 | _pt.rpn & 0xc07f | ev.value1 << 7;
		if (_pt.rpn >= 0xc000)  // Received MSB,LSB
		{
			_pt.paramNumber = _pt.rpn & 0x3fff;
			if (_pt.paramNumber > 0x7f)_pt.paramNumber = 0x100;
		}
	}
	private void _ccFunc_78(MidiEvent ev)
	{
		// All Sound Off --->
		int _ch = ev.channel;
		if (voiceParts[_ch].mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			if ( _vu.partNumber == _ch)_vu.soundOff();
		}
	}
	private void _ccFunc_79(MidiEvent ev)
	{
		// Reset All Controller --->
		int _ch = ev.channel;
		VoicePart _pt = voiceParts[_ch];
		_pt.resetAllControlle();
		if (_pt.mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)
			{
				_vu.updPartPitchBend(_pt.pitchBend);
				_vu.updPartModulation(_pt.ccValues[1]);
				_vu.updPartExpression(_pt.ccValues[11]);
			}
		}
	}
	private void _ccFunc_7B(MidiEvent ev)
	{
		// All Note Off --->
		int _ch = ev.channel;
		if (voiceParts[_ch].mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			// KeyOn -->
			if (((_vu.state & 0x010) != 0) && _vu.partNumber == _ch)_vu.keyOff();
		}
	}
// RPN,NRPN -----
	/*
	- RPN,NRPN -
		         [FE DCBA987 6543210]
		value -> [dc bbbbbbb aaaaaaa]
		a = RPN,NRPN (LSB)
		b = RPN,NRPN (MSB)
		c = Flag (Received LSB)
		d = Flag (Received MSB)
			
		paramNumber -> RPN[0-127], NRPN[128-255], nullRPN[256]
		value <= 100h [100h : null RPN] 
		
		_RPN_FUNC
		Function Detail : void (int ch, int param, int mask, int inc)
		
	*/
	//protected final Function_RPN<Integer,Integer,Integer,Integer>[] _RPN_FUNC = new Function_RPN[257];
	protected final Function_RPN[] _RPN_FUNC = new Function_RPN[257];

	protected boolean _rpnCtrlFlag  = true;
	protected boolean _nrpnCtrlFlag = false;
	public boolean getRPNCtrlFlag(){ return _rpnCtrlFlag; }
	public void setRPNCtrlFlag(boolean flag)
	{
		_rpnCtrlFlag = flag;
		for (int i = 0; i < 16; i++) voiceParts[i].resetRPNControlle();
	}
	public boolean getNRPNCtrlFlag(){ return _nrpnCtrlFlag; }
	public void setNRPNCtrlFlag(boolean flag)
	{
		_nrpnCtrlFlag = flag;
		for (int i = 0; i < 16; i++) voiceParts[i].resetRPNControlle();
	}
	private void _rpnFunc_UB(int ch, int param, int mask, int inc){}
	private void _rpnFunc_Null(int ch, int param, int mask, int inc){}
	// bendSens
	private void _rpnFunc_00(int ch, int param, int mask, int inc)
	{
		if (mask == 0x3f80) return;  // DataEntry LSB
		// DataEntry MSB, DataInc, DataDec ->
		int _val = (voiceParts[ch].bendSens << 7) & mask | param;
		_val = (_val >> 7) + inc;
		if		(_val <   0) _val = 0; 
		else if (_val > 127) _val = 127; 
		voiceParts[ch].bendSens = _val;
		if (voiceParts[ch].mode == 0) return;
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitBase _vu = _voiceUnits[i];
			if (_vu.partNumber == ch)_vu.updPartBendSens(_val);
		}
	}
	// masterTune (FineTune)
	private void _rpnFunc_01(int ch, int param, int mask, int inc)
	{
		//fineTune = (fineTune & mask | param) + inc;
		//v = (fineTune - 8192) * 0.08333333333333333;  // 0.083 = 8192(1cotv)/12(tones)/8192(+-)
	}
	// masterTune (CoarseTune)
	private void _rpnFunc_02(int ch, int param, int mask, int inc)
	{
		// if (mask != 0x7f) return;  // Data Entry MSB only 
		//coarseTune = param >> 7;
		//v = (coarseTune - 64) * 682.6666666666666;  // 682.6 = 8192(1cotv)/12(tones)
	}
	// masterVolume
	private void _nrpnFunc_00(int ch, int param, int mask, int inc)
	{
		if (mask != 0x7f) return;
		// DataEntry MSB ->
		setMasterVolume(_masterVolumeTable[param >> 7]);
	}
	// Default Part Mode Change
	private void _nrpnFunc_20(int ch, int param, int mask, int inc)
	{
		if (mask != 0x7f) return;
		// DataEntry MSB ->
		setDefaultPartMode(ch, param >> 7);
	}
	// Part Mode Change
	private void _nrpnFunc_21(int ch, int param, int mask, int inc)
	{
		if (mask != 0x7f) return;
		// DataEntry MSB ->
		setPartMode(ch, param >> 7);
	}
// Interface - IfGrayBoxModule -----
	public void updateSampleRate(float sampleRate)
	{
		if (sampleRate == _sampleRate) return;
		setSampleRate(sampleRate);
	}
	public void updateAudioChannels(int ch){}
	public void updateProcessBlockSize(int size){}
	
	public void resetModule()
	{
		allSoundStop();
		allPartReset();
	}
	public void initiModule(){ initialize(); }
	public void connectModule(GrayBox host){ setSampleRate(host.getSampleRate()); }
	public void disconnectModule(){ allSoundStop(); }
	public void updateBypassStatus(boolean bypass){ allSoundStop(); }
	
	public void startProcess(){ allSoundStop(); }
	public void stopProcess(){ allSoundStop(); }
	public void suspendProcess(){ allSoundStop(); }
	public void resumeProcess(){ allSoundStop(); }
	public void process(ProcessData procData)
	{
		MidiEvent _events = procData.midiEventList.next;
		float[] _outL = procData.audioOut[0], _outR = procData.audioOut[1];
		int _startFrame = procData.offsetFrames, _pos = _startFrame;
		int _mxp = _maxPolyphony;
		VoiceUnitBase[] _vu = _voiceUnits;
		// Prepare Process -->
		for (int i = 0; i < _mxp; i++)_vu[i].startFrame = _startFrame;
		// Process Events -->
		while (_events != null)
		{
			if ((_events.midiPorts & midiInPorts) == 0){ _events = _events.next; continue; }
			switch (_events.type)
			{
				case 0x80:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						for (int i = 0; i < _mxp; i++)if (_vu[i].state != 0)_vu[i].process.func(_outL, _outR, _pos);
					}
					_noteOff(_events);
					break;
				case 0x90:
					if (voiceParts[_events.channel].mode == 0) break;
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						for (int i = 0; i < _mxp; i++)if (_vu[i].state != 0)_vu[i].process.func(_outL, _outR, _pos);
					}
					if (_events.value1 > 0)_noteOn(_events);
					else _noteOff(_events);
					break;
				case 0xA0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						for (int i = 0; i < _mxp; i++)if (_vu[i].state != 0)_vu[i].process.func(_outL, _outR, _pos);
					}
					_keyAfterTouch(_events);
					break;
				case 0xB0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						for (int i = 0; i < _mxp; i++)if (_vu[i].state != 0)_vu[i].process.func(_outL, _outR, _pos);
					}
					_CC_FUNC[_events.value0].func(_events);
					break;
				case 0xC0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						for (int i = 0; i < _mxp; i++)if (_vu[i].state != 0)_vu[i].process.func(_outL, _outR, _pos);
					}
					voiceParts[_events.channel].programChange = _events.value0;
					break;
				case 0xD0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						for (int i = 0; i < _mxp; i++)if (_vu[i].state != 0)_vu[i].process.func(_outL, _outR, _pos);
					}
					_channelAfterTouch(_events);
					break;
				case 0xE0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						for (int i = 0; i < _mxp; i++)if (_vu[i].state != 0)_vu[i].process.func(_outL, _outR, _pos);
					}
					_pitchBendChange(_events);
					break;
				case 0xF0:
					if (!receiveSysEx) break;
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						for (int i = 0; i < _mxp; i++)if (_vu[i].state != 0)_vu[i].process.func(_outL, _outR, _pos);
					}
					_readSysExData(_events);
					break;
			}
			_events = _events.next;
		}
		// End Process -->
		int _endFrame = _startFrame + procData.sampleFrames;
		for (int i = 0; i < _mxp; i++)if (_vu[i].state != 0)_vu[i].process.func(_outL, _outR, _endFrame);
	}
}