/*! ********************************************
 * SynthesizerFM4.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth;

import sora.graybox.MidiEvent;
// 2op FM(PM) Synthesis
public final class SynthesizerFM4 extends MultiTimberSynthBase
{
private class VoiceUnitFM4 extends VoiceUnitBase
{
// Tables
	// Log2 Volume Table
	private float[] _LOG2V_TABLE = null;
	private int[] _LOG2V_TABLE_I = null;
	// Pow2 Table
	private final int _POW2T_OCTV_BITS  = SynthData.POW2T_OCTV_BITS;
	private final int _POW2T_OCTV_MASK  = SynthData.POW2T_OCTV_MASK;
	private int[] _POW2_TABLE      = null;
	private int[] _POW2T_RS_TABLE  = null;
	private int[] _POW2T_KEY_TABLE = null;
	// EG
	private final int _EG_PRAM_OCTV_LS = SynthData.POW2T_OCTV_BITS - 3;
	private final int _EG_CNT_FP       = SynthData.EG_CNT_FP;
	private final int _EG_MAX_CNT      = SynthData.EG_MAX_CNT;
	private final int _EG_CNT_ENV_ATK  = SynthData.EG_CNT_ENV_ATK;
	private final int _EG_CNT_ENV_BTM  = SynthData.EG_CNT_ENV_BTM;
	private final int _EG_SUSTAIN_BTM  = SynthData.EG_SUSTAIN_BTM;
	private int[] _EG_TOTAL_LEVEL_TABLE = null;
	private int[] _EG_LEVEL_TABLE_UP    = null;
	private int[] _EG_LEVEL_TABLE_DW    = null;
	private int[] _EG_KEY_TABLE         = null;
	private int[] _EG_POW2T_RS_TABLE    = null;
	private int[] _EG_CNT_GATE_TABLE    = null;
	// Wave
	private final int _WAVE_TABLE_MASK = SynthesizerFM4.WAVE_TABLE_SIZE - 1;
	// PG
	private final int _PG_PHASE_FP      = SynthesizerFM4.PG_PHASE_FP;
	private final int _PG_PHASE_FP_MASK = (1 << _PG_PHASE_FP) - 1;  // after the fixed-point
	private final int _PG_PHASE_MASK    = (SynthesizerFM4.WAVE_TABLE_SIZE << _PG_PHASE_FP) - 1;
	// FM
	private int[] _MUL_TABLE = null;
	// LFO
	private int[][] _LFO_WAVE_TABLES = null;
	private final int _LFO_TIMER_INIT_VALUE  = SynthData.LFO_TIMER_INIT_VALUE;
	private final int _LFO_WAVE_TABLE_MASK   = SynthData.LFO_WAVE_TABLE_SIZE - 1;
	// Synth
	VoicePart[] VOICE_PARTS = null;
	int[][] WAVE_TABLES = null;
	TonePresetFM4[] TONE_PRESETS = null;
	SampleData[] SAMPLE_DATAS = null;
	SamplerPreset[] SAMPLER_PRESETS = null;
	int pgSamplePow2tIdx  = 0;  // PG PhaseStep (Pow2Table Index Offset) - SamplingRate
	int pgKeyPow2tIdx     = 0;  // PG PhaseStep (Pow2Table Index Offset) - ReferenceKey (NoteNumber 0)
	int egPow2tIdx        = 0;  // EG PhaseStep (Pow2Table Index Offset)
	int egPhaseStepSndOff = 0;  // EG PhaseStep - SoundOff
	int lfoTimerCoef      = 0;  // Coefficient Value - LFO Phase Update Timer
	// Pitch
	private int _ptBend     = 0;   // pitchBend
	private int _ptBendSens = 0;   // pitchBend Sensitivity
	// LFO
	private int _lfoPhase      = 0;  // LFO Phase
	private int _lfoTimer      = 0;  // LFO Phase Update Timer
	private int _lfoTimerStep  = 0;  // Update Timer Step
// Operator 0,1 ---
	private int[] _lfoWaveTable = null;
	private int _pgBend = 0;  // v = _ptBend * _ptBendSens
	private int _fdbk   = 0;
	
// Operator 0 ---
	// Wave
	private int[] _waveTable_0 = null;
	// LFO
	private int _lfoMod_0      = 0;  // Modulation Depth
	private int _lfoModSens_0  = 0;  // Modulation Sensitivity
	private int _lfoOutPM_0    = 0;  // output - phase modulation
	private int _lfoOutAM_0    = 0;  // output - amplitude modulation
	// PG
	private int _pgKey_0       = 0;
	private int _pgPhase_0     = 0;
	private int _pgPhaseStep_0 = 0;
	// EG
	/*
	- Envelop Generator State -
	[ 5674 3210 ] 8Bits Flag Status (0 = deActive)
	[ ---- RSDA ] "-" = Reserved Bits
		A = Attack  stage
		D = Decay   stage
		S = Sustain stage
		R = Release stage
	*/
	private int _egState_0       = 0;
	private int[] _egLevelTable_0= null;
	private int _egCount_0       = 0;
	private int _egCountStep_0   = 0;
	//private int _egCountStepAR = 0;
	private int _egCountStepDR_0 = 0;
	private int _egCountStepRR_0 = 0;
	private int _egCntEnvSus_0   = 0;
	private int _egCntEnvBtm_0   = 0;
	private int _egGate_0        = 0;
	private int _egSL_0          = 0;
	private int _egTL_0          = 0;
	private int _egTLv_0         = 0;
// Operator 1 ---
	// Wave
	private int[] _waveTable_1 = null;
	// LFO
	private int _lfoMod_1      = 0;
	private int _lfoModSens_1  = 0;
	private int _lfoOutPM_1    = 0;
	private int _lfoOutAM_1    = 0;
	// PG
	private int _pgKey_1       = 0;
	private int _pgPhase_1     = 0;
	private int _pgPhaseStep_1 = 0;
	// EG
	private int _egState_1       = 0;
	private int[] _egLevelTable_1= null;
	private int _egCount_1       = 0;
	private int _egCountStep_1   = 0;
	//private int _egCountStepAR_1 = 0;
	private int _egCountStepDR_1 = 0;
	private int _egCountStepRR_1 = 0;
	private int _egCntEnvSus_1   = 0;
	private int _egCntEnvBtm_1   = 0;
	private int _egGate_1        = 0;
	private int _egSL_1          = 0;
	private int _egTL_1          = 0;
	private int _egTLv_1         = 0;
// Operator 2 ---
	// Wave
	private int[] _waveTable_2 = null;
	// LFO
	private int _lfoMod_2      = 0;
	private int _lfoModSens_2  = 0;
	private int _lfoOutPM_2    = 0;
	private int _lfoOutAM_2    = 0;
	// PG
	private int _pgKey_2       = 0;
	private int _pgPhase_2     = 0;
	private int _pgPhaseStep_2 = 0;
	// EG
	private int _egState_2       = 0;
	private int[] _egLevelTable_2= null;
	private int _egCount_2       = 0;
	private int _egCountStep_2   = 0;
	//private int _egCountStepAR_2 = 0;
	private int _egCountStepDR_2 = 0;
	private int _egCountStepRR_2 = 0;
	private int _egCntEnvSus_2   = 0;
	private int _egCntEnvBtm_2   = 0;
	private int _egGate_2        = 0;
	private int _egSL_2          = 0;
	private int _egTL_2          = 0;
	private int _egTLv_2         = 0;
// Operator 3 ---
	// Wave
	private int[] _waveTable_3 = null;
	// LFO
	private int _lfoMod_3      = 0;
	private int _lfoModSens_3  = 0;
	private int _lfoOutPM_3    = 0;
	private int _lfoOutAM_3    = 0;
	// PG
	private int _pgKey_3       = 0;
	private int _pgPhase_3     = 0;
	private int _pgPhaseStep_3 = 0;
	// EG
	private int _egState_3       = 0;
	private int[] _egLevelTable_3= null;
	private int _egCount_3       = 0;
	private int _egCountStep_3   = 0;
	//private int _egCountStepAR_3 = 0;
	private int _egCountStepDR_3 = 0;
	private int _egCountStepRR_3 = 0;
	private int _egCntEnvSus_3   = 0;
	private int _egCntEnvBtm_3   = 0;
	private int _egGate_3        = 0;
	private int _egSL_3          = 0;
	private int _egTL_3          = 0;
	private int _egTLv_3         = 0;
// Sampler
	private int _pgPhaseAFP  = 0;  // PG Phase (After the Fixed Point)
	private int _plbkE       = 0;  // PlayBack End Frame   [plbkE <= channelPerSamples]
	private int _loopE       = 0;  // Loop End Frame       [loopE <= plbkE]
	private int _loopF       = 0;  // Loop Frames          [val = loopE - lpStart]
	private int _loopFk      = 0;  // Loop Frames          * after -> keyOff
	
	@Override
	void keyOff()
	{
		// KeyOn -->
		if ((state & 0x0010) != 0)
		{
		// OP0 is Active -->
			if ((state & 1) != 0)
			{
				if (_egState_0 == 1)
				{
					_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
					if (_egCount_0 >= _EG_MAX_CNT)_egCount_0 = 0;
					else _egCount_0 = _EG_LEVEL_TABLE_UP[_egCount_0 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				_egGate_0 = _egCntEnvBtm_0;
				_egCountStep_0 = _egCountStepRR_0;
				_egState_0 = 8;
				if (_egCount_0 > _egCntEnvBtm_0)
				{ 
					if ((state & 0xe) != 0){ state &= 0xfe; _egState_0 = 0; }
					else { killProcess(); return; }
				}
				// sampler ---
				_loopE = _plbkE;
				_loopF = _loopFk;
			}
		// OP1 is Active -->
			if ((state & 2) != 0)
			{
				if (_egState_1 == 1)
				{
					_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
					if (_egCount_1 >= _EG_MAX_CNT)_egCount_1 = 0;
					else _egCount_1 = _EG_LEVEL_TABLE_UP[_egCount_1 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				_egGate_1 = _egCntEnvBtm_1;
				_egCountStep_1 = _egCountStepRR_1;
				_egState_1 = 8;
				if (_egCount_1 > _egCntEnvBtm_1)
				{ 
					if ((state & 0xd) != 0){ state &= 0xfd; _egState_1 = 0; }
					else { killProcess(); return; }
				}
			}
			// OP2 is Active -->
			if ((state & 4) != 0)
			{
				if (_egState_2 == 1)
				{
					_egLevelTable_2 = _EG_LEVEL_TABLE_DW;
					if (_egCount_2 >= _EG_MAX_CNT)_egCount_2 = 0;
					else _egCount_2 = _EG_LEVEL_TABLE_UP[_egCount_2 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				_egGate_2 = _egCntEnvBtm_2;
				_egCountStep_2 = _egCountStepRR_2;
				_egState_2 = 8;
				if (_egCount_2 > _egCntEnvBtm_2)
				{
					if ((state & 0xb) != 0) { state &= 0xfb; _egState_2 = 0; }
					else { killProcess(); return; }
				}
			}
			// OP3 is Active -->
			if ((state & 8) != 0)
			{
				if (_egState_3 == 1)
				{
					_egLevelTable_3 = _EG_LEVEL_TABLE_DW;
					if (_egCount_3 >= _EG_MAX_CNT)_egCount_3 = 0;
					else _egCount_3 = _EG_LEVEL_TABLE_UP[_egCount_3 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				_egGate_3 = _egCntEnvBtm_3;
				_egCountStep_3 = _egCountStepRR_3;
				_egState_3 = 8;
				if (_egCount_3 > _egCntEnvBtm_3)
				{
					if ((state & 0x7) != 0) { state &= 0xf7; _egState_3 = 0; }
					else { killProcess(); return; }
				}
			}
			state = 0x0020 | state & 0xf;
		}
	}
	@Override
	void soundOff()
	{
		// KeyOn || KeyOff --->
		if ((state & 0x0030) != 0)
		{
		// OP0 Active --->
			if ((state & 1) != 0)
			{
				if (_egState_0 == 1)
				{
					_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
					if (_egCount_0 >= _EG_MAX_CNT)_egCount_0 = 0;
					else _egCount_0 = _EG_LEVEL_TABLE_UP[_egCount_0 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				if (_egCount_0 > _egCntEnvBtm_0)
				{ 
					if ((state & 0xe) != 0){ state &= 0xfe; }
					else { killProcess(); return; }
				}
				_egGate_0 = _egCntEnvBtm_0;
				_egCountStep_0 = egPhaseStepSndOff;
				_egState_0 = 0;
				// sampler ---
				_loopE = _plbkE;
				_loopF = _loopFk;
			}
		// OP1 Active --->
			if ((state & 2) != 0)
			{
				if (_egState_1 == 1)
				{
					_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
					if (_egCount_1 >= _EG_MAX_CNT)_egCount_1 = 0;
					else _egCount_1 = _EG_LEVEL_TABLE_UP[_egCount_1 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				if (_egCount_1 > _egCntEnvBtm_1)
				{ 
					if ((state & 0xd) != 0){ state &= 0xfd; }
					else { killProcess(); return; }
				}
				_egGate_1 = _egCntEnvBtm_1;
				_egCountStep_1 = egPhaseStepSndOff;
				_egState_1 = 0;
			}
		// OP2 Active --->
			if ((state & 4) != 0)
			{
				if (_egState_2 == 1)
				{
					_egLevelTable_2 = _EG_LEVEL_TABLE_DW;
					if (_egCount_2 >= _EG_MAX_CNT)_egCount_2 = 0;
					else _egCount_2 = _EG_LEVEL_TABLE_UP[_egCount_2 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				if (_egCount_2 > _egCntEnvBtm_2)
				{ 
					if ((state & 0xb) != 0){ state &= 0xfb; }
					else { killProcess(); return; }
				}
				_egGate_2 = _egCntEnvBtm_2;
				_egCountStep_2 = egPhaseStepSndOff;
				_egState_2 = 0;
				// sampler ---
				_loopE = _plbkE;
				_loopF = _loopFk;
			}
		// OP3 Active --->
			if ((state & 8) != 0)
			{
				if (_egState_3 == 1)
				{
					_egLevelTable_3 = _EG_LEVEL_TABLE_DW;
					if (_egCount_3 >= _EG_MAX_CNT)_egCount_3 = 0;
					else _egCount_3 = _EG_LEVEL_TABLE_UP[_egCount_3 >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				}
				if (_egCount_3 > _egCntEnvBtm_3)
				{
					if ((state & 0x7) != 0) { state &= 0xf7; }
					else { killProcess(); return; }
				}
				_egGate_3 = _egCntEnvBtm_3;
				_egCountStep_3 = egPhaseStepSndOff;
				_egState_3 = 0;
				// sampler ---
				_loopE = _plbkE;
				_loopF = _loopFk;
			}
			state = 0x0040 | state & 0xf;
			partNumber = -1; // SoundOff
		}
	}
	private int _shiftEGState_0()
	{
		switch (_egState_0)
		{
			// ATTACK ->
			case 1:
				_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
				_egCount_0 = 0;
				// -> DECAY
				if (_egSL_0 > 0)
				{
					_egGate_0 = _egCntEnvSus_0;
					_egCountStep_0 = _egCountStepDR_0;
					return _egState_0 = 2;
				}
				// -> SUSTAIN
				_egGate_0 = 0x7fffffff;
				_egCountStep_0 = 0;
				return _egState_0 = 4;
			// DECAY ->
			case 2:
				// -> SUSTAIN
				if ((_egTL_0 + _egSL_0) < _EG_SUSTAIN_BTM)
				{
					_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
					_egGate_0 = 0x7fffffff;
					_egCountStep_0 = 0;
					_egCount_0 = _egCntEnvSus_0;
					return _egState_0 = 4;
				}
			// DECAY,RELEASE,SoundOFF -> EG OFF
			default: return _egState_0 = 0;
		}
	}
	private int _shiftEGState_1()
	{
		switch (_egState_1)
		{
			// ATTACK ->
			case 1:
				_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
				_egCount_1 = 0;
				// -> DECAY
				if (_egSL_1 > 0)
				{
					_egGate_1 =  _egCntEnvSus_1;
					_egCountStep_1 = _egCountStepDR_1;
					return _egState_1 = 2;
				}
				// -> SUSTAIN
				_egGate_1 = 0x7fffffff;
				_egCountStep_1 = 0;
				return _egState_1 = 4;
			// DECAY ->
			case 2:
				// -> SUSTAIN
				if ((_egTL_1 + _egSL_1) < _EG_SUSTAIN_BTM)
				{
					_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
					_egGate_1 = 0x7fffffff;
					_egCountStep_1 = 0;
					_egCount_1 = _egCntEnvSus_1;
					return _egState_1 = 4;
				}
			// DECAY,RELEASE,SoundOFF -> EG OFF
			default: return _egState_1 = 0;
		}
	}
	private int _shiftEGState_2()
	{
		switch (_egState_2)
		{
			// ATTACK ->
			case 1:
				_egLevelTable_2 = _EG_LEVEL_TABLE_DW;
				_egCount_2 = 0;
				// -> DECAY
				if (_egSL_2 > 0)
				{
					_egGate_2 = _egCntEnvSus_2;
					_egCountStep_2 = _egCountStepDR_2;
					return _egState_2 = 2;
				}
				// -> SUSTAIN
				_egGate_2 = 0x7fffffff;
				_egCountStep_2 = 0;
				return _egState_2 = 4;
			// DECAY ->
			case 2:
				// -> SUSTAIN
				if ((_egTL_2 + _egSL_2) < _EG_SUSTAIN_BTM)
				{
					_egLevelTable_2 = _EG_LEVEL_TABLE_DW;
					_egGate_2 = 0x7fffffff;
					_egCountStep_2 = 0;
					_egCount_2 = _egCntEnvSus_2;
					return _egState_2 = 4;
				}
			// DECAY,RELEASE,SoundOFF -> EG OFF
			default: return _egState_2 = 0;
		}
	}
	private int _shiftEGState_3()
	{
		switch (_egState_3)
		{
			// ATTACK ->
			case 1:
				_egLevelTable_3 = _EG_LEVEL_TABLE_DW;
				_egCount_3 = 0;
				// -> DECAY
				if (_egSL_3 > 0)
				{
					_egGate_3 =  _egCntEnvSus_3;
					_egCountStep_3 = _egCountStepDR_3;
					return _egState_3 = 2;
				}
				// -> SUSTAIN
				_egGate_3 = 0x7fffffff;
				_egCountStep_3 = 0;
				return _egState_3 = 4;
			// DECAY ->
			case 2:
				// -> SUSTAIN
				if ((_egTL_3 + _egSL_3) < _EG_SUSTAIN_BTM)
				{
					_egLevelTable_3 = _EG_LEVEL_TABLE_DW;
					_egGate_3 = 0x7fffffff;
					_egCountStep_3 = 0;
					_egCount_3 = _egCntEnvSus_3;
					return _egState_3 = 4;
				}
			// DECAY,RELEASE,SoundOFF -> EG OFF
			default: return _egState_3 = 0;
		}
	}
	@Override
	void keyOn(int ch, int note, int velo)
	{
	// KeyOn Status ---
		state = 0x1f;
		partNumber = ch;
		noteNumber = note;
	// tone ---
		VoicePart _part = VOICE_PARTS[partNumber];
		TonePresetFM4 _preset;
		int _keyFlw_0 = egPow2tIdx, 
		    _keyFlw_1 = egPow2tIdx,
		    _keyFlw_2 = egPow2tIdx,
		    _keyFlw_3 = egPow2tIdx;
		if (_part.mode == 2)
		{
			_preset = TONE_PRESETS[noteNumber | 128];
			_pgKey_0 = _POW2T_KEY_TABLE[_preset.key] + _MUL_TABLE[_preset.mul_0] + (_preset.tune_0 - 8192) + pgKeyPow2tIdx;
			_pgKey_1 = _POW2T_KEY_TABLE[_preset.key] + _MUL_TABLE[_preset.mul_1] + (_preset.tune_1 - 8192) + pgKeyPow2tIdx;
			_pgKey_2 = _POW2T_KEY_TABLE[_preset.key] + _MUL_TABLE[_preset.mul_2] + (_preset.tune_2 - 8192) + pgKeyPow2tIdx;
			_pgKey_3 = _POW2T_KEY_TABLE[_preset.key] + _MUL_TABLE[_preset.mul_3] + (_preset.tune_3 - 8192) + pgKeyPow2tIdx;
			//_keyFlw = 0;
			_prPan = _preset.pan;
		}
		else
		{
			_preset = TONE_PRESETS[_part.programChange];
			_pgKey_0 = _POW2T_KEY_TABLE[noteNumber] + _MUL_TABLE[_preset.mul_0] + (_preset.tune_0 - 8192) + pgKeyPow2tIdx;
			_keyFlw_0 += _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart_0] << _preset.kf_0;
			_pgKey_1 = _POW2T_KEY_TABLE[noteNumber] + _MUL_TABLE[_preset.mul_1] + (_preset.tune_1 - 8192) + pgKeyPow2tIdx;
			_keyFlw_1 += _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart_1] << _preset.kf_1;
			_pgKey_2 = _POW2T_KEY_TABLE[noteNumber] + _MUL_TABLE[_preset.mul_2] + (_preset.tune_2 - 8192) + pgKeyPow2tIdx;
			_keyFlw_2 += _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart_2] << _preset.kf_2;
			_pgKey_3 = _POW2T_KEY_TABLE[noteNumber] + _MUL_TABLE[_preset.mul_3] + (_preset.tune_3 - 8192) + pgKeyPow2tIdx;
			_keyFlw_3 += _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart_3] << _preset.kf_3;
			_prPan = 64;
		}
	// LFO ---
		_lfoWaveTable = _LFO_WAVE_TABLES[_preset.lfoWave];
		_lfoTimerStep = _preset.lfoFreq * lfoTimerCoef;
		_lfoTimer = _LFO_TIMER_INIT_VALUE;
		_lfoPhase = _preset.strc >> 3;  // ~8 AM
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// pitch ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
	// process ---
		process = _PROC_PTRS[_preset.strc];
		_fdbk = _EG_TOTAL_LEVEL_TABLE[_preset.fl];
// Operator 0 --->
		_waveTable_0 = WAVE_TABLES[_preset.wave_0];
	// LFO ---
		_lfoModSens_0 = _preset.lms_0;
		_lfoMod_0 = _part.ccValues[1] * _lfoModSens_0;
		_lfoOutPM_0 = 0;
		_lfoOutAM_0 = 0;
	// PG ---
		int v = _pgKey_0 + _pgBend;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase_0 = 0;
	// EG ---
		_egTL_0 = _preset.tl_0 + (127 - velo >> 7 - _preset.tlVelCtrl_0);
		if (_egTL_0 > 127)_egTL_0 = 0;
		_egTLv_0 = _EG_TOTAL_LEVEL_TABLE[_egTL_0];
		_egCntEnvBtm_0 = _EG_CNT_GATE_TABLE[128 - _egTL_0];
		_egSL_0 = _preset.sl_0;
		_egCntEnvSus_0 = _EG_CNT_GATE_TABLE[_egSL_0];
		if (_egCntEnvSus_0 > _egCntEnvBtm_0)_egCntEnvSus_0 = _egCntEnvBtm_0;
		v = (_preset.dr_0 << _EG_PRAM_OCTV_LS) + _keyFlw_0;
		_egCountStepDR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr_0 << _EG_PRAM_OCTV_LS) + _keyFlw_0;
		_egCountStepRR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar_0 < 127) {
			_egCount_0 = _EG_CNT_ENV_ATK;
			_egLevelTable_0 = _EG_LEVEL_TABLE_UP;
			_egGate_0 = _EG_MAX_CNT;
			v = (_preset.ar_0 << _EG_PRAM_OCTV_LS) + _keyFlw_0;
			_egCountStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState_0 = 1;
		} else {
			_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
			_egCount_0 = 0;
		// -> DECAY
			if (_egSL_0 > 0) {
				_egGate_0 = _egCntEnvSus_0;
				_egCountStep_0 = _egCountStepDR_0;
				_egState_0 = 2;
		// -> SUSTAIN
			} else {
				_egGate_0 = 0x7fffffff;
				_egCountStep_0 = 0;
				_egState_0 = 4; }
		}
// Operator 1 --->
		_waveTable_1 = WAVE_TABLES[_preset.wave_1];
	// LFO ---
		_lfoModSens_1 = _preset.lms_1;
		_lfoMod_1 = _part.ccValues[1] * _lfoModSens_1;
		_lfoOutPM_1 = 0;
		_lfoOutAM_1 = 0;
	// PG ---
		v = _pgKey_1 + _pgBend;
		_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase_1 = 0;
	// EG ---
		_egTL_1 = _preset.tl_1 + (127 - velo >> 7 - _preset.tlVelCtrl_1);
		if (_egTL_1 > 127)_egTL_1 = 0;
		_egTLv_1 = _EG_TOTAL_LEVEL_TABLE[_egTL_1];
		_egCntEnvBtm_1 = _EG_CNT_GATE_TABLE[128 - _egTL_1];
		_egSL_1 = _preset.sl_1;
		_egCntEnvSus_1 = _EG_CNT_GATE_TABLE[_egSL_1];
		if (_egCntEnvSus_1 > _egCntEnvBtm_1)_egCntEnvSus_1 = _egCntEnvBtm_1;
		v = (_preset.dr_1 << _EG_PRAM_OCTV_LS) + _keyFlw_1;
		_egCountStepDR_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr_1 << _EG_PRAM_OCTV_LS) + _keyFlw_1;
		_egCountStepRR_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar_1 < 127) {
			_egCount_1 = _EG_CNT_ENV_ATK;
			_egLevelTable_1 = _EG_LEVEL_TABLE_UP;
			_egGate_1 = _EG_MAX_CNT;
			v = (_preset.ar_1 << _EG_PRAM_OCTV_LS) + _keyFlw_1;
			_egCountStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState_1 = 1;
		} else {
			_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
			_egCount_1 = 0;
		// -> DECAY
			if (_egSL_1 > 0) {
				_egGate_1 = _egCntEnvSus_1;
				_egCountStep_1 = _egCountStepDR_1;
				_egState_1 = 2;
		// -> SUSTAIN
			} else {
				_egGate_1 = 0x7fffffff;
				_egCountStep_1 = 0;
				_egState_1 = 4; }
		}
// Operator 2 --->
		_waveTable_2 = WAVE_TABLES[_preset.wave_2];
	// LFO ---
		_lfoModSens_2 = _preset.lms_2;
		_lfoMod_2 = _part.ccValues[1] * _lfoModSens_2;
		_lfoOutPM_2 = 0;
		_lfoOutAM_2 = 0;
	// PG ---
		v = _pgKey_2 + _pgBend;
		_pgPhaseStep_2 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase_2 = 0;
	// EG ---
		_egTL_2 = _preset.tl_2 + (127 - velo >> 7 - _preset.tlVelCtrl_2);
		if (_egTL_2 > 127)_egTL_2 = 0;
		_egTLv_2 = _EG_TOTAL_LEVEL_TABLE[_egTL_2];
		_egCntEnvBtm_2 = _EG_CNT_GATE_TABLE[128 - _egTL_2];
		_egSL_2 = _preset.sl_2;
		_egCntEnvSus_2 = _EG_CNT_GATE_TABLE[_egSL_2];
		if (_egCntEnvSus_2 > _egCntEnvBtm_2)_egCntEnvSus_2 = _egCntEnvBtm_2;
		v = (_preset.dr_2 << _EG_PRAM_OCTV_LS) + _keyFlw_2;
		_egCountStepDR_2 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr_2 << _EG_PRAM_OCTV_LS) + _keyFlw_2;
		_egCountStepRR_2 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar_2 < 127) {
			_egCount_2 = _EG_CNT_ENV_ATK;
			_egLevelTable_2 = _EG_LEVEL_TABLE_UP;
			_egGate_2 = _EG_MAX_CNT;
			v = (_preset.ar_2 << _EG_PRAM_OCTV_LS) + _keyFlw_2;
			_egCountStep_2 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState_2 = 1;
		} else {
			_egLevelTable_2 = _EG_LEVEL_TABLE_DW;
			_egCount_2 = 0;
		// -> DECAY
			if (_egSL_2 > 0) {
				_egGate_2 = _egCntEnvSus_2;
				_egCountStep_2 = _egCountStepDR_2;
				_egState_2 = 2;
		// -> SUSTAIN
			} else {
				_egGate_2 = 0x7fffffff;
				_egCountStep_2 = 0;
				_egState_2 = 4; }
		}
// Operator 3 --->
		_waveTable_3 = WAVE_TABLES[_preset.wave_3];
	// LFO ---
		_lfoModSens_3 = _preset.lms_3;
		_lfoMod_3 = _part.ccValues[1] * _lfoModSens_3;
		_lfoOutPM_3 = 0;
		_lfoOutAM_3 = 0;
	// PG ---
		v = _pgKey_3 + _pgBend;
		_pgPhaseStep_3 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase_3 = 0;
	// EG ---
		_egTL_3 = _preset.tl_3 + (127 - velo >> 7 - _preset.tlVelCtrl_3);
		if (_egTL_3 > 127)_egTL_3 = 0;
		_egTLv_3 = _EG_TOTAL_LEVEL_TABLE[_egTL_3];
		_egCntEnvBtm_3 = _EG_CNT_GATE_TABLE[128 - _egTL_3];
		_egSL_3 = _preset.sl_3;
		_egCntEnvSus_3 = _EG_CNT_GATE_TABLE[_egSL_3];
		if (_egCntEnvSus_3 > _egCntEnvBtm_3)_egCntEnvSus_3 = _egCntEnvBtm_3;
		v = (_preset.dr_3 << _EG_PRAM_OCTV_LS) + _keyFlw_3;
		_egCountStepDR_3 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr_3 << _EG_PRAM_OCTV_LS) + _keyFlw_3;
		_egCountStepRR_3 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar_3 < 127) {
			_egCount_3 = _EG_CNT_ENV_ATK;
			_egLevelTable_3 = _EG_LEVEL_TABLE_UP;
			_egGate_3 = _EG_MAX_CNT;
			v = (_preset.ar_3 << _EG_PRAM_OCTV_LS) + _keyFlw_3;
			_egCountStep_3 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState_3 = 1;
		} else {
			_egLevelTable_3 = _EG_LEVEL_TABLE_DW;
			_egCount_3 = 0;
		// -> DECAY
			if (_egSL_3 > 0) {
				_egGate_3 = _egCntEnvSus_3;
				_egCountStep_3 = _egCountStepDR_3;
				_egState_3 = 2;
		// -> SUSTAIN
			} else {
				_egGate_3 = 0x7fffffff;
				_egCountStep_3 = 0;
				_egState_3 = 4; }
		}
	}
	void keyOnSampler(int ch, int note, int velo)
	{
	// KeyOn Status ---
		state = 0x11;
		partNumber = ch;
		noteNumber = note;
	// tone ---
		VoicePart _part = VOICE_PARTS[partNumber];
		SamplerPreset _preset = SAMPLER_PRESETS[noteNumber];
		SampleData _sample = SAMPLE_DATAS[_preset.smplNum];
		if (_sample.sampleData == null){ killProcess(); return; }
		_waveTable_0 = _sample.sampleData;
		_plbkE = _sample.plbkE;
		_loopE = _sample.loopE;
		_loopF = _sample.loopF;
		_loopFk = _sample.loopFk;
	// LFO ---
		_lfoOutPM_0 = 0;
		_lfoOutAM_0 = 0;
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_prPan = _preset.pan;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// PG ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		_pgKey_0 = (_preset.octv - 8 << 13) + (_preset.tune - 8192) + _sample.pgSample + pgSamplePow2tIdx;
		int v = _pgKey_0 + _pgBend;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase_0 = _sample.plbkS;
		_pgPhaseAFP = 0;
	// EG ---
		_egTL_0 = _preset.tl;
		_egTLv_0 = _EG_TOTAL_LEVEL_TABLE[_egTL_0];
		_egCntEnvBtm_0 = _EG_CNT_GATE_TABLE[128 - _egTL_0];
		_egSL_0 = _preset.sl;
		_egCntEnvSus_0 = _EG_CNT_GATE_TABLE[_egSL_0];
		if (_egCntEnvSus_0 > _egCntEnvBtm_0)_egCntEnvSus_0 = _egCntEnvBtm_0;
		v = (_preset.dr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepDR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepRR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar < 127) {
			_egCount_0 = _EG_CNT_ENV_ATK;
			_egLevelTable_0 = _EG_LEVEL_TABLE_UP;
			_egGate_0 = _EG_MAX_CNT;
			v = (_preset.ar << _EG_PRAM_OCTV_LS) + egPow2tIdx;
			_egCountStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState_0 = 1;
		} else {
			_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
			_egCount_0 = 0;
		// -> DECAY
			if (_egSL_0 > 0) {
				_egGate_0 = _egCntEnvSus_0;
				_egCountStep_0 = _egCountStepDR_0;
				_egState_0 = 2;
		// -> SUSTAIN
			} else {
				_egGate_0 = 0x7fffffff;
				_egCountStep_0 = 0;
				_egState_0 = 4; }
		}
	// process ---
		process = (_sample.channelCount == 2)? this::_procSmpl2ch: this::_procSmpl1ch;
	}
	@Override
	void updPartBendSens(int val)
	{
		_ptBendSens = _POW2T_KEY_TABLE[val];
		_pgBend = _ptBend * _ptBendSens >> 13;
		int v;
		// Inst op0 --->
		if ((state & 1) != 0)
		{
			v = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
		// Inst op1 --->
		if ((state & 2) != 0)
		{
			v = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
		// Inst op2 --->
		if ((state & 4) != 0)
		{
			v = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
		// Inst op3 --->
		if ((state & 8) != 0)
		{
			v = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
	}
	@Override
	void updPartPitchBend(int val)
	{
		_ptBend = val - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		int v;
		// Inst op0 --->
		if ((state & 1) != 0)
		{
			v = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
		// Inst op1 --->
		if ((state & 2) != 0)
		{
			v = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
		// Inst op2 --->
		if ((state & 4) != 0)
		{
			v = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
		// Inst op3 --->
		if ((state & 8) != 0)
		{
			v = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		}
	}
	@Override
	void updPartModulation(int val)
	{
		_lfoMod_0 = val * _lfoModSens_0;
		_lfoMod_1 = val * _lfoModSens_1;
		_lfoMod_2 = val * _lfoModSens_2;
		_lfoMod_3 = val * _lfoModSens_3;
	}
	void updateLfoFrequency(int val)
	{
		_lfoTimerStep = val * lfoTimerCoef;
	}
	void setup(SynthData synthData)
	{
		// volume Table
		_mainVolTable = synthData.getVOLUME_TABLE_0dB();
		_exprVolTable = synthData.getVOLUME_TABLE_0dB();
		_veloVolTable = synthData.getVOLUME_TABLE_0dB();
		_panTable     = synthData.getPAN_TABLE_SIN();
		// Log2 Volume Table
		_LOG2V_TABLE    = synthData.getLOG2V_TABLE();
		_LOG2V_TABLE_I  = synthData.getLOG2V_TABLE_I();
		// Pow2 Table
		_POW2_TABLE      = synthData.getPOW2_TABLE();
		_POW2T_RS_TABLE  = synthData.getPOW2T_RS_TABLE();
		_POW2T_KEY_TABLE = synthData.getPOW2T_KEY_TABLE();
		// EG
		_EG_TOTAL_LEVEL_TABLE = synthData.getEG_TOTAL_LEVEL_TABLE();
		_EG_LEVEL_TABLE_UP    = synthData.getEG_LEVEL_TABLE_UP();
		_EG_LEVEL_TABLE_DW    = synthData.getEG_LEVEL_TABLE_DW();
		_EG_KEY_TABLE         = synthData.getEG_KEY_TABLE();
		_EG_POW2T_RS_TABLE    = synthData.getEG_POW2T_RS_TABLE();
		_EG_CNT_GATE_TABLE    = synthData.getEG_CNT_GATE_TABLE();
		// etc
		_MUL_TABLE       = synthData.getMUL_TABLE();
		_LFO_WAVE_TABLES = synthData.getLFO_WAVE_TABLES();
	}
// Constructor
	VoiceUnitFM4()
	{
		super();
		_PROC_PTRS[0] = this::_proc0;
		_PROC_PTRS[1] = this::_proc1;
		_PROC_PTRS[2] = this::_proc2;
		_PROC_PTRS[3] = this::_proc3;
		_PROC_PTRS[4] = this::_proc4;
		_PROC_PTRS[5] = this::_proc5;
		_PROC_PTRS[6] = this::_proc6;
		_PROC_PTRS[7] = this::_proc7;
		_PROC_PTRS[8] = this::_proc8;
		_PROC_PTRS[9] = this::_proc9;
		_PROC_PTRS[10] = this::_procA;
		_PROC_PTRS[11] = this::_procB;
		_PROC_PTRS[12] = this::_procC;
		_PROC_PTRS[13] = this::_procD;
		_PROC_PTRS[14] = this::_procE;
		_PROC_PTRS[15] = this::_procF;
	}
// process ---
	// pointer array (process Function)
	private final Function_PRC[] _PROC_PTRS = new Function_PRC[16];

	private void _proc0(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				// op0
				_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op0
				_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
				x = _pgKey_2 + _pgBend + _lfoOutPM_2;
				_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
				x = _pgKey_3 + _pgBend + _lfoOutPM_3;
				_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0){ state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				// OP 0 out --->
				n = _log2vT[p];
				// counter --->
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else n = 0.0f;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
				// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
				// pg update --->
				x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
				// pg update --->
				x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else
					{
						if ((state & 0x7) != 0) { state &= 0xf7; _egCount_3 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
				// pg update --->
				x = _wt_3[_pgPhase_3 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _proc1(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				// op0
				_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op0
				_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
				x = _pgKey_2 + _pgBend + _lfoOutPM_2;
				_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
				x = _pgKey_3 + _pgBend + _lfoOutPM_3;
				_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[p];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
			// OP 1 out ----->
				n = _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else n = 0.0f;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
				// pg update --->
				x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else
					{
						if ((state & 0x7) != 0) { state &= 0xf7; _egCount_3 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
				// pg update --->
				x = _wt_3[_pgPhase_3 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _proc2(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				// op0
				_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op0
				_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
				x = _pgKey_2 + _pgBend + _lfoOutPM_2;
				_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
				x = _pgKey_3 + _pgBend + _lfoOutPM_3;
				_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[p];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
				// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
				// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else n = 0.0f;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
				// pg update --->
				p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_2[p] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else
					{
						if ((state & 0x7) != 0) { state &= 0xf7; _egCount_3 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _proc3(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				// op0
				_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op0
				_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
				x = _pgKey_2 + _pgBend + _lfoOutPM_2;
				_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
				x = _pgKey_3 + _pgBend + _lfoOutPM_3;
				_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[x];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
				// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
				// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else n = 0.0f;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
				// pg update --->
				x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
				// OP 0 out --->
				o = _log2vTI[x];
				// counter --->	
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else
					{
						if ((state & 0x7) != 0) { state &= 0xf7; _egCount_3 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _proc4(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				// op0
				_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op0
				_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
				x = _pgKey_2 + _pgBend + _lfoOutPM_2;
				_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
				x = _pgKey_3 + _pgBend + _lfoOutPM_3;
				_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[x];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
			// OP 1 out ----->
				o = _log2vTI[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
				// pg update --->
				x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				o += _log2vTI[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else { killProcess(); return; }
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _proc5(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				// op0
				_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op0
				_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
				x = _pgKey_2 + _pgBend + _lfoOutPM_2;
				_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
				x = _pgKey_3 + _pgBend + _lfoOutPM_3;
				_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 1 --->
			if ((state & 2) != 0)
			{
				// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
				// pg update --->
				p = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				o = _log2vTI[p];
				// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
				// pg update --->
				p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_2[p] + x;
				// OP 1 out ----->
				o = _log2vTI[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o += _log2vTI[x];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else { killProcess(); return; }
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _proc6(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				// op0
				_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op0
				_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
				x = _pgKey_2 + _pgBend + _lfoOutPM_2;
				_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
				x = _pgKey_3 + _pgBend + _lfoOutPM_3;
				_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[x];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
				o += _log2vTI[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
				// pg update --->
				p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_2[p] + x;
				// OP 1 out ----->
				o = _log2vTI[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else { killProcess(); return; }
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _proc7(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				// op0
				_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
				x = _pgKey_0 + _pgBend + _lfoOutPM_0;
				_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
				x = _pgKey_1 + _pgBend + _lfoOutPM_1;
				_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op0
				_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
				x = _pgKey_2 + _pgBend + _lfoOutPM_2;
				_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// op1
				_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
				x = _pgKey_3 + _pgBend + _lfoOutPM_3;
				_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
				// counter
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[p];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				p = _wt_1[p] + x;
			// OP 1 out ----->
				o = _log2vTI[p];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
				// pg update --->
				p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				p = _wt_2[p] + x;
				// OP 1 out ----->
				o = _log2vTI[p];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else { killProcess(); return; }
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				p = _wt_3[p] + x;
				// OP 1 out ----->
				n = _log2vT[p];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}

	private void _proc8(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				x = (p * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (p * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				x = (p * _lfoMod_2) >> 11;
				_lfoOutAM_2 = x & 0x7ffffffe;
				x = (p * _lfoMod_3) >> 11;
				_lfoOutAM_3 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0){ state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				// OP 0 out --->
				n = _log2vT[p];
				// counter --->
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			} else n = 0.0f;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
				// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
				// pg update --->
				x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
				// pg update --->
				x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else
					{
						if ((state & 0x7) != 0) { state &= 0xf7; _egCount_3 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
				// pg update --->
				x = _wt_3[_pgPhase_3 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _proc9(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				x = (p * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (p * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				x = (p * _lfoMod_2) >> 11;
				_lfoOutAM_2 = x & 0x7ffffffe;
				x = (p * _lfoMod_3) >> 11;
				_lfoOutAM_3 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[p];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
			// OP 1 out ----->
				n = _log2vT[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else n = 0.0f;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
				// pg update --->
				x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else
					{
						if ((state & 0x7) != 0) { state &= 0xf7; _egCount_3 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
				// pg update --->
				x = _wt_3[_pgPhase_3 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _procA(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				x = (p * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (p * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				x = (p * _lfoMod_2) >> 11;
				_lfoOutAM_2 = x & 0x7ffffffe;
				x = (p * _lfoMod_3) >> 11;
				_lfoOutAM_3 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[p];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
				// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
				// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else n = 0.0f;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
				// pg update --->
				p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_2[p] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else
					{
						if ((state & 0x7) != 0) { state &= 0xf7; _egCount_3 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _procB(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				x = (p * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (p * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				x = (p * _lfoMod_2) >> 11;
				_lfoOutAM_2 = x & 0x7ffffffe;
				x = (p * _lfoMod_3) >> 11;
				_lfoOutAM_3 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[x];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
				// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
				// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else n = 0.0f;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
				// pg update --->
				x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
				// OP 0 out --->
				o = _log2vTI[x];
				// counter --->	
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else
					{
						if ((state & 0x7) != 0) { state &= 0xf7; _egCount_3 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n += _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _procC(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				x = (p * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (p * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				x = (p * _lfoMod_2) >> 11;
				_lfoOutAM_2 = x & 0x7ffffffe;
				x = (p * _lfoMod_3) >> 11;
				_lfoOutAM_3 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[x];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_1[p] + x;
			// OP 1 out ----->
				o = _log2vTI[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
				// pg update --->
				x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				o += _log2vTI[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else { killProcess(); return; }
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _procD(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				x = (p * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (p * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				x = (p * _lfoMod_2) >> 11;
				_lfoOutAM_2 = x & 0x7ffffffe;
				x = (p * _lfoMod_3) >> 11;
				_lfoOutAM_3 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 1 --->
			if ((state & 2) != 0)
			{
				// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
				// pg update --->
				p = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
				// OP 1 out ----->
				o = _log2vTI[p];
				// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
				// pg update --->
				p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_2[p] + x;
				// OP 1 out ----->
				o = _log2vTI[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o += _log2vTI[x];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else { killProcess(); return; }
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _procE(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				x = (p * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (p * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				x = (p * _lfoMod_2) >> 11;
				_lfoOutAM_2 = x & 0x7ffffffe;
				x = (p * _lfoMod_3) >> 11;
				_lfoOutAM_3 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[x];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
				x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
				o += _log2vTI[x];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
				// pg update --->
				p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_2[p] + x;
				// OP 1 out ----->
				o = _log2vTI[x];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else { killProcess(); return; }
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				x = _wt_3[p] + x;
				// OP 1 out ----->
				n = _log2vT[x];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
	private void _procF(float[] outL, float[] outR, int end)
	{
		int[] _et_0 = _egLevelTable_0, _et_1 = _egLevelTable_1,
		      _et_2 = _egLevelTable_2, _et_3 = _egLevelTable_3;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _log2vTI = _LOG2V_TABLE_I; 
		final int[] _wt_0 = _waveTable_0, _wt_1 = _waveTable_1,
		            _wt_2 = _waveTable_2, _wt_3 = _waveTable_3;
		int x, p, f, o; float n = 0.0f;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				p = _lfoWaveTable[_lfoPhase];
				x = (p * _lfoMod_0) >> 11;
				_lfoOutAM_0 = x & 0x7ffffffe;
				x = (p * _lfoMod_1) >> 11;
				_lfoOutAM_1 = x & 0x7ffffffe;
				x = (p * _lfoMod_2) >> 11;
				_lfoOutAM_2 = x & 0x7ffffffe;
				x = (p * _lfoMod_3) >> 11;
				_lfoOutAM_3 = x & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// Operator 0 --->
			if ((state & 1) != 0)
			{
				// eg update --->
				if (_egCount_0 >= _egGate_0)
				{
					if ((_shiftEGState_0() != 0))_et_0 = _egLevelTable_0;
					else
					{
						if ((state & 0xe) != 0) { state &= 0xfe; _egCount_0 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
				// pg update --->
				p = _pgPhase_0 >> _PG_PHASE_FP;
				// FeedBack -->
				f = _log2vTI[_wt_0[p] + _fdbk];
				// OP 0 out --->
				p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
				o = _log2vTI[p];
				// counter --->	
				_egCount_0 += _egCountStep_0;
				_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 1 --->
			if ((state & 2) != 0)
			{
			// eg update --->
				if (_egCount_1 >= _egGate_1)
				{
					if ((_shiftEGState_1() != 0))_et_1 = _egLevelTable_1;
					else
					{
						if ((state & 0xd) != 0) { state &= 0xfd; _egCount_1 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
				p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				p = _wt_1[p] + x;
			// OP 1 out ----->
				o = _log2vTI[p];
			// counter --->
				_egCount_1 += _egCountStep_1;
				_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 2 --->
			if ((state & 4) != 0)
			{
				// eg update --->
				if (_egCount_2 >= _egGate_2)
				{
					if ((_shiftEGState_2() != 0))_et_2 = _egLevelTable_2;
					else
					{
						if ((state & 0xb) != 0) { state &= 0xfb; _egCount_2 = _EG_MAX_CNT - 1; }
						else { killProcess(); return; }
					}
				}
				x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
				// pg update --->
				p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				p = _wt_2[p] + x;
				// OP 1 out ----->
				o = _log2vTI[p];
				// counter --->
				_egCount_2 += _egCountStep_2;
				_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
			}
			else o = 0;
		// Operator 3 --->
			if ((state & 8) != 0)
			{
				// eg update --->
				if (_egCount_3 >= _egGate_3)
				{
					if ((_shiftEGState_3() != 0))_et_3 = _egLevelTable_3;
					else { killProcess(); return; }
				}
				x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
				// pg update --->
				p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
				p = _wt_3[p] + x;
				// OP 1 out ----->
				n = _log2vT[p];
				// counter --->
				_egCount_3 += _egCountStep_3;
				_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
			}
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
		}
		startFrame = i;
	}
// process --- Perc
	// Sampler 1ch
	private void _procSmpl1ch(float[] outL, float[] outR, int end)
	{
		int[] _et = _egLevelTable_0;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _wt = _waveTable_0;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0() != 0)_et = _egLevelTable_0;
				else { killProcess(); return; }
			}
			int  v = _et[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;	
		// pg update --->
			if (_pgPhase_0 >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase_0 -= _loopF; } while (_pgPhase_0 >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase_0] + v;
			float n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhaseAFP += _pgPhaseStep_0;
			_pgPhase_0 += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Sampler 2ch
	private void _procSmpl2ch(float[] outL, float[] outR, int end)
	{
		int[] _et = _egLevelTable_0;
		final float[] _log2vT = _LOG2V_TABLE; 
		final int[] _wt = _waveTable_0;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState_0() != 0)_et = _egLevelTable_0;
				else { killProcess(); return; }
			}
			int v = _et[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;	
		// pg update --->
			if (_pgPhase_0 >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase_0 -= _loopF; } while (_pgPhase_0 >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			int p = _pgPhase_0 << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhaseAFP += _pgPhaseStep_0;
			_pgPhase_0 += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
}
// Factory -----
	@Override
	protected void _recreateVU(int maxPoly)
	{
		// Recreate Array
		_voiceUnits = new VoiceUnitBase[maxPoly];
		_voiceUnitExt = new VoiceUnitFM4[maxPoly];
		// Reset VoiceUnit
		for (int i = 0; i < maxPoly; i++)
		{
			_voiceUnitExt[i] = new VoiceUnitFM4();
			VoiceUnitFM4 _vu = _voiceUnitExt[i];
			_voiceUnits[i] = _vu;
			_vu.setup(_synthData);
			_vu.killProcess();
			_vu.VOICE_PARTS = voiceParts;
			_vu.WAVE_TABLES = waveTables;
			_vu.SAMPLE_DATAS = sampleDatas;
			_vu.TONE_PRESETS = tonePresets;
			_vu.SAMPLER_PRESETS = samplerPresets;
			_vu.setMasterVolume(_masterVolume);
		}
		// set Property
		_maxPolyphony = maxPoly;
	}
	@Override
	protected void _resetKeyParams()
	{
		final double _sampleRateD = (double)_sampleRate;
		// LFO
		double _lfoWaveSize = SynthData.LFO_WAVE_TABLE_SIZE / 2.0;
			// Max Frame Rate - LFO Phase Step
		double _lfoMaxFrameRate = _sampleRateD / SynthData.LFO_MAX_FREQ / _lfoWaveSize;
		if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
		int _lfoTimerCoef =  (int)(SynthData.LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		// EG
		double _rate0Frames = SynthData.EG_REFERENCE_TIME * SynthData.LOG2VT_DYN_BITS * _sampleRateD;
		int _egPow2tIdx = SynthData.toPow2tIndexValue(SynthData.EG_MAX_CNT / _rate0Frames);
		double _sndOffFrames = SynthData.EG_SOUND_OFF_TIME / SynthData.LOG2VT_DYN_BITS * _sampleRateD;
		int _egSndOff = (int)(SynthData.EG_MAX_CNT / _sndOffFrames + 0.5);
		// PG PCM
		int _smpl = SynthData.toPow2tIndexValue(_sampleRateD);
		int _pgSamplePow2tIdx = SynthData.toPow2tIndexValue((double)(1 << PG_PHASE_FP)) - _smpl;
		// PG WaveTable
		int _pgKeyPow2tIdx = SynthData.toPow2tIndexValue((double)(WAVE_TABLE_SIZE << PG_PHASE_FP)) - _smpl;
		_pgKeyPow2tIdx += SynthData.toPow2tIndexValue(_referencePitch * Math.pow(2.0, -69.0 / 12.0)); // note 0 Pitch
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitFM4 _vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
			_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
			_vu.egPow2tIdx = _egPow2tIdx;
			_vu.egPhaseStepSndOff = _egSndOff;
			_vu.lfoTimerCoef = _lfoTimerCoef;
		}
	}
// Synthesizer - Data -----
	/*  
		Multiple
		MUL 0 - 1/2 pai       MUL 8 -  8 pai
		MUL 1 -   1 pai       MUL 9 -  9 pai
		MUL 2 -   2 pai       MUL10 - 10 pai
		MUL 3 -   3 pai       MUL11 - 11 pai
		MUL 4 -   4 pai       MUL12 - 12 pai
		MUL 5 -   5 pai       MUL13 - 13 pai
		MUL 6 -   6 pai       MUL14 - 14 pai
		MUL 7 -   7 pai       MUL15 - 15 pai
		* From Key Pitch
		
		Total Level
		         mod : car (pai)
		TL  0 -  64 : 1      TL 64 - 1 /   8 : 1 
		TL  8 -  32 : 1      TL 72 - 1 /  16 : 1
		TL 16 -  16 : 1      TL 80 - 1 /  32 : 1
		TL 24 -   8 : 1      TL 88 - 1 /  64 : 1
		TL 32 -   4 : 1      TL 96 - 1 / 128 : 1
		TL 40 -   2 : 1      TL104 - 1 / 256 : 1
		TL 48 -   1 : 1      TL112 - 1 / 512 : 1
		TL 56 - 1/2 : 1      TL120 - 1 /1024 : 1
		TL 60 - 1/4 : 1      TL128 - 1 /2048 : 1  (table size)
		* Modulation Depth
		  LOG2VT output -> +-65536(131072)(Max) : (1024)WaveTableSize
	*/
	static public final int WAVE_TABLE_SIZE_BITS = 11;                        // 16384 (DataEntry MSB LSB) / 8 (waves)
	static public final int WAVE_TABLE_SIZE      = 1 << WAVE_TABLE_SIZE_BITS; // 16384 (DataEntry MSB LSB) / 8 (waves)
	static final int PG_PHASE_FP = 16;  // 16bit fixed-point  (precision 1 / 65536)
	static final int OVER_TONES  = 16;  // for Sinewave Synthesis.
	private int[][] waveTables = null;
	private SampleData[] sampleDatas = null;
	public TonePresetFM4[] tonePresets = null;
	public SamplerPreset[] samplerPresets = null;
	private SynthData _synthData = null;
	public void getWaveForm(int id, float[] buffer)
	{
		if (buffer.length < WAVE_TABLE_SIZE) return;
		// ---
		int[] _table = waveTables[id];
		final float[] _log2vt = _synthData.getLOG2V_TABLE();
		for (int i = 0; i < WAVE_TABLE_SIZE; i++) buffer[i] = _log2vt[_table[i]];
	}
	public void setWaveForm(int id, float[] buffer)
	{
		if (buffer.length < WAVE_TABLE_SIZE) return;
		final double _LOGE = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387);
		final float _BOADER = 1.0f / (1 << SynthData.LOG2VT_DYN_BITS);
		final int _BOTTOM = SynthData.LOG2VT_SIZE;
		int[] _table = waveTables[id];
		for (int i = 0; i < WAVE_TABLE_SIZE; i++)
		{
			float v = buffer[i];
			if      (v >  _BOADER) _table[i] =  (int)(Math.log((double)( v)) * _LOGE + 0.5) << 1;
			else if (v < -_BOADER) _table[i] = ((int)(Math.log((double)(-v)) * _LOGE + 0.5) << 1) + 1;
			else _table[i] = _BOTTOM;
		}
	}
	public SampleData getSampleData(int id){ return sampleDatas[id]; }
	public void setSampleData(int id, SampleData sampleData, boolean deepCopy)
	{
		allSoundStop();
		sampleDatas[id].copy(sampleData, deepCopy);
	}
	public void loadPCMSamples(int id, float[] wave, int ch)
	{
		allSoundStop();
		sampleDatas[id].loadPCMSamples(wave, ch);
	}
	public void loadPCMData(int id, byte[] dat, int position, int sampleFrames, int ch, int bitRate, boolean sign, boolean bigEndian)
	{
		allSoundStop();
		sampleDatas[id].loadPCMData(dat, position, sampleFrames, ch, bitRate, sign, bigEndian);
	}
	public void clearSampleData(int id)
	{
		allSoundStop();
		sampleDatas[id].clear();
	}
	@Override
	public void factoryDataReset()
	{
		if (_voiceUnitExt != null) allSoundStop();
		
		final double _LOGE = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387);
		final double _BOADER = 1.0 / (1 << SynthData.LOG2VT_DYN_BITS);
		final int _BOTTOM = SynthData.LOG2VT_SIZE;
	// Sample
		final int _SIZE = WAVE_TABLE_SIZE, _HALF = _SIZE >> 1, _QTER = _SIZE >> 2;
		final double _SIZEd = (double)_SIZE, _HALFd = (double)_HALF, _QTERd = (double)_QTER;
		final double _PID2 = 3.141592653589793 * 2.0;
		double[] _sl = new double[OVER_TONES];
		double[] _sw = new double[OVER_TONES];
		if (waveTables == null)waveTables = new int[8][];
		for (int i = 0; i < 8; i++) if (waveTables[i] == null) waveTables[i] = new int[_SIZE];
		for (int i = 0, _iv = 0; i < _HALF; i++)
		{
		// sine [ nu ]
			double _nv = Math.sin((i / _SIZEd) * _PID2);
			if      (_nv >  _BOADER) _iv =  (int)(Math.log( _nv) * _LOGE + 0.5) << 1;
			else if (_nv < -_BOADER) _iv = ((int)(Math.log(-_nv) * _LOGE + 0.5) << 1) + 1;
			else _iv = _BOTTOM;
			waveTables[0][i] = _iv;
			waveTables[0][_HALF + i] = _iv + 1;
		// sine [ n- ]
			waveTables[1][i] = _iv;
			waveTables[1][_HALF + i] = _BOTTOM;
		// sine [ nn ]
			waveTables[2][i] = _iv; 
			waveTables[2][_HALF + i] = _iv;
			if	((i & 1) == 0){ 
		// sine [ nu-- ]
				waveTables[3][i >> 1] = _iv; 
				waveTables[3][_QTER + (i >> 1)] = _iv + 1;
		// sine [ nn-- ]
				waveTables[4][i >> 1] = _iv; 
				waveTables[4][_QTER + (i >> 1)] = _iv; }
			waveTables[3][_HALF + i] = _BOTTOM;
			waveTables[4][_HALF + i] = _BOTTOM;
		}
	// triangle ----
		for (int i = 0, j = 1; i < OVER_TONES; i++, j+=2)_sw[i] = _SIZEd / j;
		for (int i = 0, j = 1; i < OVER_TONES; i++, j+=2)_sl[i] = ((i & 1) != 0)? -1.0 / (j * j) : 1.0 / (j * j);
		double _den = 0.0;
		for (int i = 0; i < OVER_TONES; i++)_den += Math.sin(_QTERd / _sw[i] * _PID2) * _sl[i];
		_den = 1.0 / _den;
		for (int i = 0, _iv = 0; i < _SIZE; i++)
		{
			double _nv = 0.0;
			for (int j = 0; j < OVER_TONES; j++)_nv += Math.sin((i / _sw[j]) * _PID2) * _sl[j];
			_nv *= _den;
			if      (_nv >  _BOADER) _iv =  (int)(Math.log( _nv) * _LOGE + 0.5) << 1;
			else if (_nv < -_BOADER) _iv = ((int)(Math.log(-_nv) * _LOGE + 0.5) << 1) + 1;
			else _iv = _BOTTOM;
			waveTables[5][i] = _iv;
		}
	// square ----
		for (int i = 0, j = 1; i < OVER_TONES; i++, j+=2)_sl[i] = 1.0 / j;
		final double  _cQTR = _QTERd / OVER_TONES;
		_den = 0.0;
		for (int i = 0; i < OVER_TONES; i++)_den += Math.sin(_cQTR / _sw[i] * _PID2) * _sl[i];
		_den = 1.0 / _den;
		for (int i = 0, _iv = 0; i < _SIZE; i++)
		{
			double _nv = 0.0;
			for (int j = 0; j < OVER_TONES; j++)_nv += Math.sin((i / _sw[j]) * _PID2) * _sl[j];
			_nv *= _den;
			if      (_nv >  _BOADER) _iv =  (int)(Math.log( _nv) * _LOGE + 0.5) << 1;
			else if (_nv < -_BOADER) _iv = ((int)(Math.log(-_nv) * _LOGE + 0.5) << 1) + 1;
			else _iv = _BOTTOM;
			waveTables[6][i] = _iv;
		}
	// saw ----
		for (int i = 0, j = 1; i < OVER_TONES; i++, j++)_sw[i] = _SIZEd / j;
		for (int i = 0, j = 1; i < OVER_TONES; i++, j++)_sl[i] = 1.0 / j;
		double _c = _HALFd;
		for (int i = 0, j = 1; i < OVER_TONES; i++, j++) _c /= 1.0 + 1.0 / j;
		_c = Math.round(_c + 0.5);
		_den = 0.0;
		for (int j = 0; j < OVER_TONES; j++)_den += Math.sin(_c / _sw[j] * _PID2) * _sl[j];
		_den = 1.0 / _den;
		for (int i = 0, _iv = 0; i < _SIZE; i++)
		{
			double _nv = 0.0;
			for (int j = 0; j < OVER_TONES; j++)_nv += Math.sin((i / _sw[j]) * _PID2) * _sl[j];
			_nv *= _den;
			if      (_nv >  _BOADER) _iv =  (int)(Math.log( _nv) * _LOGE + 0.5) << 1;
			else if (_nv < -_BOADER) _iv = ((int)(Math.log(-_nv) * _LOGE + 0.5) << 1) + 1;
			else _iv = _BOTTOM;
			waveTables[7][i] = _iv;
		}
	// Preset
		if (tonePresets == null)tonePresets = new TonePresetFM4[256];
		for (int i = 0; i < 256; i++)
		{
			if (tonePresets[i] == null)tonePresets[i] = new TonePresetFM4();
			tonePresets[i].reset();
		}
	// Sampler
		if (sampleDatas == null)sampleDatas = new SampleData[128];
		if (samplerPresets == null)samplerPresets = new SamplerPreset[128];
		for (int i = 0; i < 128; i++)
		{
			if (sampleDatas[i] == null)sampleDatas[i] = new SampleData();
			else sampleDatas[i].clear();
			if (samplerPresets[i] == null)samplerPresets[i] = new SamplerPreset();
			else samplerPresets[i].reset();
		}
	}
// Synthesizer - main -----
	private VoiceUnitFM4[] _voiceUnitExt = null;
// Constructor
	public SynthesizerFM4(){ this(32, 440.0f, 1.0f, true); }
	public SynthesizerFM4(int maxPoly, float refPitch, float masVol, boolean privateTableData)
	{
	// super ===
		super(maxPoly, refPitch, masVol);
	// SynthesizerFM4 ===
		_modelID = 0x22;
		_maxPMN = 3;
		// create
		_synthData = (privateTableData)? new SynthData() : SynthData.getInstance();
		factoryDataReset();
		_recreateVU(_maxPolyphony);
		_resetKeyParams();
	}
// MidiEvent ---
	@Override
	protected void _noteOn(MidiEvent ev)
	{
		VoicePart _pt = voiceParts[ev.channel];
		final int _start = _nton;
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitFM4 _vu = _voiceUnitExt[_nton];
			if (_vu.state == 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
			
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitFM4 _vu = _voiceUnitExt[_nton];
			// (KeyOff || SoundOff) -->
			if ((_vu.state & 0x60) != 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
		
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitFM4 _vu = _voiceUnitExt[_nton];
			// KeyOn -->
			if ((_vu.state & 0x10) != 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
	}
	@Override
	protected void _readSysExData(MidiEvent ev)
	{
		final byte[] _data = ev.data;
		final int _dataEnd = ev.dataSize - 1;  // minus F7
		if (_dataEnd < 8)return;
		int p = 0, v = 0;
		if ((_data[p++]&0xff) != _manufacID) return;
		if ((_data[p++]&0xff) != _deiceID) return;
		if ((_data[p++]&0xff) != _modelID) return;
		int b = _data[p++]&0xff;
		switch (b)	// Address 1
		{
		// Common
			case 0x00:
				b = (_data[p++]&0xff)<<8 | (_data[p++]&0xff);
				switch (b)	// Address 2,3
				{
					case 0x0000 : setMasterVolume(_masterVolumeTable[(_data[p++]&127)]); if (p >= _dataEnd) break;
					case 0x0001 : setReferencePitch(440.0f + ((_data[p++]&127) - 64) * 0.5f); break;
				}
				break;
		// Part
			case 0x02:
				v = _data[p++] & 15;	// Address 2 [ CH ]
				b = _data[p++]&0x7f;
				switch (b)	// Address 3
				{
					case 0x00 : setDefaultPartMode(v, _data[p++]); if (p >= _dataEnd) break;
					case 0x01 : setPartMode(v, _data[p++]); break;
				}
				break;
		// Inst (Drum map)
			case 0x05: v = 128;
		// Inst (Program map)
			case 0x04:
				TonePresetFM4 _preset = tonePresets[v | (_data[p++]&127)];	// Address 2 [ ID ]
				b = (_data[p++]&0xff);
				switch (b)	// Address 3
				{
					case 0x00 : _preset.toneName[0]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x01 : _preset.toneName[1]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x02 : _preset.toneName[2]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x03 : _preset.toneName[3]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x04 : _preset.toneName[4]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x05 : _preset.toneName[5]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x06 : _preset.toneName[6]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x07 : _preset.toneName[7]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x08 : _preset.toneName[8]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x09 : _preset.toneName[9]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0A : _preset.toneName[10] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0B : _preset.toneName[11] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0C : _preset.toneName[12] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0D : _preset.toneName[13] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0E : _preset.toneName[14] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0F : _preset.toneName[15] = (char)_data[p++]; if (p >= _dataEnd) break;
					
					case 0x10 : p++; if (p >= _dataEnd) break;
					case 0x11 : p++; if (p >= _dataEnd) break;
					case 0x12 : p++; if (p >= _dataEnd) break;
					case 0x13 : p++; if (p >= _dataEnd) break;
					case 0x14 : p++; if (p >= _dataEnd) break;
					case 0x15 : p++; if (p >= _dataEnd) break;
					case 0x16 : p++; if (p >= _dataEnd) break;
					case 0x17 : p++; if (p >= _dataEnd) break;
					case 0x18 : p++; if (p >= _dataEnd) break;
					case 0x19 : _preset.key     = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1A : _preset.pan     = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1B : _preset.lfoWave = _data[p++] & 3;   if (p >= _dataEnd) break;
					case 0x1C : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x1D : _preset.lfoFreq = _preset.lfoFreq & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x1E : _preset.fl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1F : _preset.strc    = _data[p++] & 7;   if (p >= _dataEnd) break;
					
					case 0x20 : p++; if (p >= _dataEnd) break;
					case 0x21 : p++; if (p >= _dataEnd) break;
					case 0x22 : p++; if (p >= _dataEnd) break;
					case 0x23 : _preset.tlVelCtrl_0 = _data[p++] & 7;   if (p >= _dataEnd) break;
					case 0x24 : _preset.mul_0       = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x25 : _preset.tune_0      = _preset.tune_0 & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x26 : _preset.tune_0      = _preset.tune_0 & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x27 : _preset.lms_0       = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x28 : _preset.wave_0      = _data[p++] & 7;   if (p >= _dataEnd) break;
					case 0x29 : _preset.tl_0        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x2A : _preset.ar_0        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x2B : _preset.dr_0        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x2C : _preset.sl_0        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x2D : _preset.rr_0        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x2E : _preset.kf_0        = _data[p++] & 3;   if (p >= _dataEnd) break;
					case 0x2F : _preset.kfStart_0   = _data[p++] & 127; if (p >= _dataEnd) break;
					
					case 0x30 : p++; if (p >= _dataEnd) break;
					case 0x31 : p++; if (p >= _dataEnd) break;
					case 0x32 : p++; if (p >= _dataEnd) break;
					case 0x33 : _preset.tlVelCtrl_1 = _data[p++] & 7;   if (p >= _dataEnd) break;
					case 0x34 : _preset.mul_1       = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x35 : _preset.tune_1      = _preset.tune_1 & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x36 : _preset.tune_1      = _preset.tune_1 & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x37 : _preset.lms_1       = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x38 : _preset.wave_1      = _data[p++] & 7;   if (p >= _dataEnd) break;
					case 0x39 : _preset.tl_1        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x3A : _preset.ar_1        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x3B : _preset.dr_1        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x3C : _preset.sl_1        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x3D : _preset.rr_1        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x3E : _preset.kf_1        = _data[p++] & 3;   if (p >= _dataEnd) break;
					case 0x3F : _preset.kfStart_1   = _data[p++] & 127; if (p >= _dataEnd) break;
					
					case 0x40 : p++; if (p >= _dataEnd) break;
					case 0x41 : p++; if (p >= _dataEnd) break;
					case 0x42 : p++; if (p >= _dataEnd) break;
					case 0x43 : _preset.tlVelCtrl_2 = _data[p++] & 7;   if (p >= _dataEnd) break;
					case 0x44 : _preset.mul_2       = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x45 : _preset.tune_2      = _preset.tune_2 & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x46 : _preset.tune_2      = _preset.tune_2 & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x47 : _preset.lms_2       = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x48 : _preset.wave_2      = _data[p++] & 7;   if (p >= _dataEnd) break;
					case 0x49 : _preset.tl_2        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x4A : _preset.ar_2        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x4B : _preset.dr_2        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x4C : _preset.sl_2        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x4D : _preset.rr_2        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x4E : _preset.kf_2        = _data[p++] & 3;   if (p >= _dataEnd) break;
					case 0x4F : _preset.kfStart_2   = _data[p++] & 127; if (p >= _dataEnd) break;
					
					case 0x50 : p++; if (p >= _dataEnd) break;
					case 0x51 : p++; if (p >= _dataEnd) break;
					case 0x52 : p++; if (p >= _dataEnd) break;
					case 0x53 : _preset.tlVelCtrl_3 = _data[p++] & 7;   if (p >= _dataEnd) break;
					case 0x54 : _preset.mul_3       = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x55 : _preset.tune_3      = _preset.tune_3 & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x56 : _preset.tune_3      = _preset.tune_3 & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x57 : _preset.lms_3       = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x58 : _preset.wave_3      = _data[p++] & 7;   if (p >= _dataEnd) break;
					case 0x59 : _preset.tl_3        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x5A : _preset.ar_3        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x5B : _preset.dr_3        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x5C : _preset.sl_3        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x5D : _preset.rr_3        = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x5E : _preset.kf_3        = _data[p++] & 3;   if (p >= _dataEnd) break;
					case 0x5F : _preset.kfStart_3   = _data[p++] & 127; break;
				}
				break;
		// Percussion
			case 0x06:
				SamplerPreset _smplPrst = samplerPresets[(_data[p++]&127)];	// Address 2 [ ID ]
				b = (_data[p++]&0xff);
				switch (b)	// Address 3
				{
					case 0x00 : _smplPrst.toneName[0]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x01 : _smplPrst.toneName[1]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x02 : _smplPrst.toneName[2]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x03 : _smplPrst.toneName[3]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x04 : _smplPrst.toneName[4]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x05 : _smplPrst.toneName[5]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x06 : _smplPrst.toneName[6]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x07 : _smplPrst.toneName[7]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x08 : _smplPrst.toneName[8]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x09 : _smplPrst.toneName[9]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0A : _smplPrst.toneName[10] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0B : _smplPrst.toneName[11] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0C : _smplPrst.toneName[12] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0D : _smplPrst.toneName[13] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0E : _smplPrst.toneName[14] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0F : _smplPrst.toneName[15] = (char)_data[p++]; if (p >= _dataEnd) break;
				
					case 0x10 : _smplPrst.smplNum = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x11 : _smplPrst.pan     = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x12 : _smplPrst.octv    = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x13 : _smplPrst.tune    = _smplPrst.tune & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x14 : _smplPrst.tune    = _smplPrst.tune & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x15 : _smplPrst.tl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x16 : _smplPrst.ar      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x17 : _smplPrst.dr      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x18 : _smplPrst.sl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x19 : _smplPrst.rr      = _data[p++] & 127; break;
				}
				break;
		}
	}
}